
/* Attention!
 * To maintain compliance with Nordic Semiconductor ASA's Bluetooth profile
 * qualification listings, this section of source code must not be modified.
 */
#include "sdk_common.h"
#if NRF_MODULE_ENABLED(BLE_POS)
#include "ble_pos.h"
#include <string.h>
#include "ble_l2cap.h"
#include "ble_srv_common.h"


#define OPCODE_LENGTH 1                                                              /**< Length of opcode inside Pulseoxymeter Measurement packet. */
#define HANDLE_LENGTH 2                                                              /**< Length of handle inside Pulseoxymeter Measurement packet. */
#define MAX_POM_LEN      (NRF_SDH_BLE_GATT_MAX_MTU_SIZE - OPCODE_LENGTH - HANDLE_LENGTH) /**< Maximum size of a transmitted Pulseoxymeter Measurement. */

#define INITIAL_VALUE_POM                       0                                    /**< Initial Pulseoxymeter Measurement value. */
#define INITIAL_FEATURES_POM                    0                                    /**< Initial Features value. */


// Pulseoxymeter Measurement flag bits
#define POM_FLAG_MASK_SPO2PR_FAST_INCLUDED          					(0x01 << 0)                           /**< Pulseoxymeter Value Format bit. */
#define POM_FLAG_MASK_SPO2PR_SLOW_INCLUDED          					(0x01 << 1)                           /**< Sensor Contact Detected bit. */
#define POM_FLAG_MASK_MEASURMENT_STATUS_INCLUDED    					(0x01 << 2)                           /**< Sensor Contact Supported bit. */
#define POM_FLAG_MASK_DEVICE_SENSOR_INCLUDED        					(0x01 << 3)                           /**< Energy Expended Status bit. Feature Not Supported */
#define POM_FLAG_MASK_PULSE_AMPLITUDE_INCLUDED      					(0x01 << 4)

#define POM_MEASUREMENT_FLAG_MASK_CALIBRATION_ONGOING     		(0x01 << 12)

#define FEATURE_FLAG_MASK_MEASUREMENT_INCLUDED     						(0x01 << 0)
#define FEATURE_FLAG_MASK_DEVICE_SENSOR_INCLUDED							(0x01 << 1)

#define FEATURE__MEASUREMENT_FLAG_MASK_CAL_ONGOING_INCLUDED   (0x01 << 12)

/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_pos       Pulseoxymeter Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_pos_t * p_pos, ble_evt_t const * p_ble_evt)
{
    p_pos->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}


/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_pos       Pulseoxymeter Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_pos_t * p_pos, ble_evt_t const * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_pos->conn_handle = BLE_CONN_HANDLE_INVALID;
}



/**@brief Function for handling write events to the Pulseoxymeter Measurement characteristic.
 *
 * @param[in]   p_pos         Pulseoxymeter Service structure.
 * @param[in]   p_evt_write   Write event received from the BLE stack.
 */
static void on_pom_c_cccd_write(ble_pos_t * p_pos, ble_gatts_evt_write_t const * p_evt_write)
{
    if (p_evt_write->len == 2)
    {
        // CCCD written, update notification state
        if (p_pos->evt_handler != NULL)
        {
            ble_pos_evt_t evt;

            if (ble_srv_is_notification_enabled(p_evt_write->data))
            {
                evt.evt_type = BLE_POS_EVT_NOTIFICATION_ENABLED;
            }
            else
            {
                evt.evt_type = BLE_POS_EVT_NOTIFICATION_DISABLED;
            }

            p_pos->evt_handler(p_pos, &evt);
        }
    }
}

/**@brief Function for handling the Write event.
 *
 * @param[in]   p_pos       Pulseoxymeter Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_pos_t * p_pos, ble_evt_t const * p_ble_evt)
{
    ble_gatts_evt_write_t const * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

    if (p_evt_write->handle == p_pos->pom_c_handles.cccd_handle)
    {
        on_pom_c_cccd_write(p_pos, p_evt_write);
    }
}


void ble_pos_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context)
{
    ble_pos_t * p_pos = (ble_pos_t *) p_context;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_pos, p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_pos, p_ble_evt);
            break;

        case BLE_GATTS_EVT_WRITE:
            on_write(p_pos, p_ble_evt);
            break;

        default:
            // No implementation needed.
            break;
    }
}


/**@brief Function for encoding a Pulseoxymeter Measurement continuous.
 *
 * @param[in]   p_pos              Pulseoxymeter Service structure.
 * @param[in]   pulseoxymeter_value         Measurement to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static uint8_t pom_c_encode(ble_pos_t * p_pos, ble_pos_meas_t * p_pos_meas, uint8_t * p_encoded_buffer)
{
    uint8_t flags = 0;
    uint8_t len   = 1;
//    int     i;
		uint16_t encoded_spo2;
		uint16_t encoded_pr;
		uint16_t measurments_flags=0;

    // Set sensor contact related flags

		flags |=POM_FLAG_MASK_MEASURMENT_STATUS_INCLUDED;// no flags  the first byte is at 0 The next bytes are float SPO2 et Float PR, the minimum functionnalities

    // Encode Pulseoxymeter measurement

		encoded_spo2 = ((p_pos_meas->spo2_value.exponent << 12) & 0xF000) |
                   ((p_pos_meas->spo2_value.mantissa) & 0x0FFF) ;

		len   += uint16_encode(encoded_spo2, &p_encoded_buffer[len]);

		encoded_pr = ((p_pos_meas->pr_value.exponent << 12) & 0xF000) |
                 ((p_pos_meas->pr_value.mantissa) & 0x0FFF) ;

		len   += uint16_encode(encoded_pr, &p_encoded_buffer[len]);

		if (p_pos->is_measurment_calibration_ongoing)
		{
				measurments_flags|=POM_MEASUREMENT_FLAG_MASK_CALIBRATION_ONGOING;
		}
		len   += uint16_encode(measurments_flags, &p_encoded_buffer[len]);
    // Add flags
    p_encoded_buffer[0] = flags;

    return len;
}

/**@brief Function for encoding the features of the pos.
 *
 * @param[in]   p_pos              Pulseoxymeter Service structure.
 * @param[in]   p_pos_features     Measurement to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static uint8_t features_encode(ble_pos_t * p_pos, ble_pos_features_t * p_pos_features, uint8_t * p_encoded_buffer)
{
    uint8_t len   = 0;
		uint16_t supported_features=0;
		uint16_t measurement_features=0;
		uint32_t device_sensor_features=0;

		// Encode supported_features
		if (p_pos->is_measurement_supported)
		{
				supported_features|= FEATURE_FLAG_MASK_MEASUREMENT_INCLUDED;
		}
		if (p_pos->is_device_sensor_supported)
		{
				supported_features|= FEATURE_FLAG_MASK_DEVICE_SENSOR_INCLUDED; //put flags
		}
		len   += uint16_encode(supported_features, &p_encoded_buffer[len]);

    // Encode measurement_status
		if (p_pos->is_caliration_ongoing_supported)
		{
				measurement_features|=FEATURE__MEASUREMENT_FLAG_MASK_CAL_ONGOING_INCLUDED;
		}
		len   += uint16_encode(measurement_features, &p_encoded_buffer[len]);

		// Encode device_sensor_support
		len   += uint24_encode(device_sensor_features, &p_encoded_buffer[len]);

    return len;
}


/**@brief Function for adding the Pulseoxymeter Measurement characteristic.
 *
 * @param[in]   p_pos        Pulseoxymeter Service structure.
 * @param[in]   p_pos_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t pulseoxymeter_value_continuous_measurement_char_add(ble_pos_t            * p_pos,
																																		const ble_pos_init_t * p_pos_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_md_t cccd_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;
    uint8_t             encoded_initial_pom[MAX_POM_LEN];

		memset(&cccd_md, 0, sizeof(cccd_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
    cccd_md.write_perm = p_pos_init->pos_pom_c_attr_md.cccd_write_perm;
    cccd_md.vloc       = BLE_GATTS_VLOC_STACK;

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.notify = 1;
    char_md.p_char_user_desc  = NULL;
    char_md.p_char_pf         = NULL;
    char_md.p_user_desc_md    = NULL;
    char_md.p_cccd_md         = &cccd_md;
    char_md.p_sccd_md         = NULL;

    BLE_UUID_BLE_ASSIGN(ble_uuid, BLE_UUID_PLX_CONTINUOUS_MEAS);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.read_perm  = p_pos_init->pos_pom_c_attr_md.read_perm;
    attr_md.write_perm = p_pos_init->pos_pom_c_attr_md.write_perm;
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = pom_c_encode(p_pos,INITIAL_VALUE_POM , encoded_initial_pom);
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = MAX_POM_LEN;
    attr_char_value.p_value   = encoded_initial_pom;

    return sd_ble_gatts_characteristic_add(p_pos->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_pos->pom_c_handles);
}


/**@brief Function for adding the Body Sensor Location characteristic.
 *
 * @param[in]   p_pos        Pulseoxymeter Service structure.
 * @param[in]   p_pos_init   Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t features_char_add(ble_pos_t * p_pos, const ble_pos_init_t * p_pos_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;
	  uint8_t             encoded_initial_features[MAX_POM_LEN];

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.read  = 1;
    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf        = NULL;
    char_md.p_user_desc_md   = NULL;
    char_md.p_cccd_md        = NULL;
    char_md.p_sccd_md        = NULL;

    BLE_UUID_BLE_ASSIGN(ble_uuid, BLE_UUID_PLX_FEATURES);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.read_perm  = p_pos_init->pos_fts_attr_md.read_perm;
    attr_md.write_perm = p_pos_init->pos_fts_attr_md.write_perm;
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));


    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = features_encode(p_pos, INITIAL_FEATURES_POM, encoded_initial_features);
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = MAX_POM_LEN;
    attr_char_value.p_value   = encoded_initial_features;

    return sd_ble_gatts_characteristic_add(p_pos->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_pos->fts_handels);

}

uint32_t ble_pos_init(ble_pos_t * p_pos, const ble_pos_init_t * p_pos_init)
{
   uint32_t   err_code;
    ble_uuid_t ble_uuid;

    // Initialize service structure
    p_pos->evt_handler                    	 = p_pos_init->evt_handler;
    p_pos->conn_handle                    	 = BLE_CONN_HANDLE_INVALID;
    p_pos->max_pom_len                 	     = MAX_POM_LEN;
		p_pos->is_device_sensor_supported  	     = false;
		p_pos->is_measurement_supported		   		 = true;
		p_pos->is_caliration_ongoing_supported   = true;
		p_pos->is_measurment_calibration_ongoing = p_pos_init->is_measurment_calibration_ongoing;

    // Add service
    BLE_UUID_BLE_ASSIGN(ble_uuid, BLE_UUID_PLX_SERVICE);

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                        &ble_uuid,
                                        &p_pos->service_handle);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    // Add Pulseoxymeter continuous measurement characteristic
    err_code = pulseoxymeter_value_continuous_measurement_char_add(p_pos, p_pos_init);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

		// Add Pulseoxymeter single measurement characteristic
    //  err_code = pulseoxymeter_value_single_measurement_char_add(p_pos, p_pos_init);
    //    if (err_code != NRF_SUCCESS)
    //  {
    //    return err_code;
    // }

		// Add features characteristic
    err_code = features_char_add(p_pos, p_pos_init);
    if (err_code != NRF_SUCCESS)
    {
				return err_code;
    }

    return NRF_SUCCESS;
}


uint32_t ble_pos_pulseoxymeter_value_measurement_send(ble_pos_t * p_pos, ble_pos_meas_t * p_pos_meas)
{
    uint32_t err_code;

    // Send value if connected and notifying
    if (p_pos->conn_handle != BLE_CONN_HANDLE_INVALID)
    {
        uint8_t                encoded_pom[MAX_POM_LEN];
        uint16_t               len;
        uint16_t               hvx_len;
        ble_gatts_hvx_params_t hvx_params;

        len     = pom_c_encode(p_pos, p_pos_meas, encoded_pom);
        hvx_len = len;

        memset(&hvx_params, 0, sizeof(hvx_params));

        hvx_params.handle = p_pos->pom_c_handles.value_handle;
        hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;
        hvx_params.offset = 0;
        hvx_params.p_len  = &hvx_len;
        hvx_params.p_data = encoded_pom;

        err_code = sd_ble_gatts_hvx(p_pos->conn_handle, &hvx_params);
        if ((err_code == NRF_SUCCESS) && (hvx_len != len))
        {
            err_code = NRF_ERROR_DATA_SIZE;
        }
    }
    else
    {
        err_code = NRF_ERROR_INVALID_STATE;
    }

    return err_code;
}

void ble_pos_measurment_calibration_ongoing_update(ble_pos_t * p_pos, bool is_measurment_calibration_ongoing)
{
    p_pos->is_measurment_calibration_ongoing = is_measurment_calibration_ongoing;
}



void ble_pos_on_gatt_evt(ble_pos_t * p_pos, nrf_ble_gatt_evt_t const * p_gatt_evt)
{
    if (    (p_pos->conn_handle == p_gatt_evt->conn_handle)
        &&  (p_gatt_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED))
    {
        p_pos->max_pom_len = p_gatt_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
    }
}
#endif // NRF_MODULE_ENABLED(BLE_pos)
